

%  Definicao das operacoes

operacao( pick_up(A),
           [na_mesa(A),not em_cima(C,A), not na_mao(X)],  % pre condicao
           [na_mao(A),not na_mesa(A)] ).                 % pos condicao

operacao( put_down(A),
           [na_mao(A)],                                   % pre condicao
           [na_mesa(A),not na_mao(A)] ).                 % pos condicao

operacao( pick_off(A,B),
           [em_cima(A,B),not em_cima(C,A), not na_mao(X)],  % pre condicao
           [na_mao(A),not em_cima(A,B)] ).                 % pos condicao

operacao( put_on(A,B),
           [na_mao(A),not em_cima(C,B)],                   % pre condicao
           [em_cima(A,B),not na_mao(A)] ).                 % pos condicao

% Metas satisfeitas

f_satisfeita([],Estado).
f_satisfeita([X|Meta],Estado) :-
        f_sat_uma(X,Estado),
        f_satisfeita(Meta,Estado).

% satisfacao de uma das metas apenas 
f_sat_uma(X,Estado) :- 
	member(X,Estado).
f_sat_uma(not X,Estado) :- 
	\+ f_sat_uma(X,Estado).

% Ontencao das diferencas estre o estado atual e as metas

f_meta_dif([],Estado,[]).
f_meta_dif([X|Meta],Estado,Mdif) :-
        f_sat_uma(X,Estado), !,
        f_meta_dif(Meta,Estado,Mdif).
f_meta_dif([not X|Meta],Estado,[not X|Mdif]) :-
        !,
        f_sat_uma(X,Estado),
        f_meta_dif(Meta,Estado,Mdif).
f_meta_dif([X|Meta],Estado,[X|Mdif]) :-
        f_meta_dif(Meta,Estado,Mdif).

% Geracao do plano para atigir as metas desejadas

f_plano(Estado,Meta,[],Estado):-
            f_satisfeita(Meta,Estado), !.
f_plano(Init,Meta,Plano,Final):-
            f_plano_uma(Init,Meta,Plano_Uma,After_Uma),
            f_plano(After_Uma,Meta,Plano_Rest,Final),      %  **
            append(Plano_Uma,Plano_Rest,Plano).

% Plano para uma das metas

f_plano_uma(Init,Meta,Plano,Final):-
            f_meta_dif(Meta,Init,Mdif),
            member(X,Mdif),
            operacao(Op,Pre,Post),
            member(X,Post),
            f_tenta_op(Init,Op,Plano,Final).

% Ipressao do que estah sendo feito

mostra_tenta(Op) :- write('tentativa '), write(Op), write(' ...'), nl.
mostra_tenta(Op) :- write('...'), write(Op), write(' - Falha!!!'), nl, fail.

% Aplicacao da operacao tentativa

f_tenta_op(Init,Op,Plano,Final) :-
            mostra_tenta(Op),
            operacao(Op,Pre,Post),
            f_plano(Init,Pre,Plano_Pre_Op,S_Pre_Op),
            f_atualiza_estado(S_Pre_Op,Post,Final),
            append(Plano_Pre_Op,[Op],Plano).

% Atualizacao dos estados em virtude das operacoes realizadas

f_atualiza_estado(Estado,[],Estado).
f_atualiza_estado(Estado,[X|Rest],Estado2) :-
            f_sat_uma(X,Estado),
            f_atualiza_estado(Estado,Rest,Estado2).
f_atualiza_estado(Estado,[not X|Rest],Estado2) :-
            !,
            remove(X,Estado,Estado1),
            f_atualiza_estado(Estado1,Rest,Estado2).
f_atualiza_estado(Estado,[X|Rest],Estado2) :-
            f_atualiza_estado([X|Estado],Rest,Estado2).

% Exemplo de empilhamento de piramides

d:- f_plano( [ na_mesa(blue_pyramid),            
                   em_cima(red_pyramid,green_box),
                   em_cima(blue_box,red_box),
                   na_mesa(green_box),
                   na_mesa(red_box)
            ],
            [ em_cima(blue_pyramid,red_box),    
                    na_mesa(red_box)
            ],
            P,E ).

% Exemplo de empilhamento de blocos cubicos

f:- f_plano( [ na_mesa(a),            
                   em_cima(c,a),
                   na_mesa(b)
            ],
            [ em_cima(a,b),    
              em_cima(b,c)
            ],
            P,E ).



