
% Exemplo de Jogo da velha com MINIMAX



inicio([b,b,b,b,b,b,b,b,b]).
profun(2).

ganhou([A,A,A,_,_,_,_,_,_],A):-A = x ; A = o.
ganhou([_,_,_,A,A,A,_,_,_],A):-A = x ; A = o.
ganhou([_,_,_,_,_,_,A,A,A],A):-A = x ; A = o.
ganhou([A,_,_,A,_,_,A,_,_],A):-A = x ; A = o.
ganhou([_,A,_,_,A,_,_,A,_],A):-A = x ; A = o.
ganhou([_,_,A,_,_,A,_,_,A],A):-A = x ; A = o.
ganhou([A,_,_,_,A,_,_,_,A],A):-A = x ; A = o.
ganhou([_,_,A,_,A,_,A,_,_],A):-A = x ; A = o.

poss([1,2,3,4,5,6,7,8,9]).
possivel(X):-
	poss(L),
	member(X,L).  %  Posicao da jogada X pertence a 1..9

velha:-  splash,
        inicio(Taboleiro),
	profun(Prof),
	jogo(Taboleiro,Prof),!.
velha:-write('deu velha ou houve um lance ilegal'),nl.

splash:-
        write('           Jogo da velha'),nl,
               write('O maximizador joga com x'),nl,
               write('O usuario joga com o'),nl,
                write('A jogada ou lance eh indicada por'),nl,
                write('posicao no taboleiro, seguida de ponto e ENTER'),nl,
                write('da forma abaixo'),nl,
                write('1'),write('2'),write('3'),
	     nl,
	     write('4'),write('5'),write('6'),
	      nl,
	      write('7'),write('8'),write('9'),
	      nl,nl,
              write('       Agora comeca o jogo'),
              nl,nl.

jogo(Taboleiro,Prof):-
	ganhou(Taboleiro,A),  %  Jogador A "o" ou "x"
	write(A),
	write(' venceu o jogo'),nl.
jogo(Taboleiro,Prof):-
	not ganhou(Taboleiro,o),
	fazmax([Taboleiro,0],R,Prof),  % Retorna a pos minimax R
	joga(x,                        % Jogador x
             Taboleiro,                     
             R,               % Posicao da nova jogada  1..9
             NovoTaboleiro),
	imprime(NovoTaboleiro),
	((not ganhou(NovoTaboleiro,x),
	write('Faca sua jogada'),
	read(NovaJogada),
	joga(o,NovoTaboleiro,NovaJogada,NovoTaboleiro2),
	jogo(NovoTaboleiro2,Prof));(
	ganhou(NovoTaboleiro,x),
	write('x venceu o jogo'),nl
	)).

imprime([A1,A2,A3,A4,A5,A6,A7,A8,A9]):-   % Taboleiro impresso por linhas
	nl,
	write(A1),write(A2),write(A3),
	nl,
	write(A4),write(A5),write(A6),
	nl,
	write(A7),write(A8),write(A9),
	nl.

fazmax(X,M,Prof):-
	maximiza(X,Prof,[R,M]).
fazmin(X,M,Prof):-
	minimiza(X,Prof,[R,M]).


maximiza(X,0,X):-!.          % Quando Prof for 0 retorna a entrada
maximiza([X,J],Prof,[X,J]):- % Quando x ganha retorna a entrada
	ganhou(X,x),!.
maximiza([X,J],Prof,M):-     % Gera um Lance Max
	bagof([L,Lance],          % Lista L gerada pela operacao joga
                                                % a partir de L com lance na pos Lance
              joga(x,X,Lance,L),  % Cada valor de Lance gera uma L
              TaboLance),             % Lista de listas [L,Lance]
	Prof2 is Prof - 1,
	minimizal(TaboLance,TaboLanceMini,Prof2),
	maiorvalor(TaboLanceMini,MaiorTaboLanceMin),
	member(MaiorTaboLanceMin,TaboLanceMini,POS),
	member(M,TaboLance,POS).

% TaboLance eh a estrutura que possui, para cada Lance possivel (1..9)
% a imagem do taboleiro e o Lance correspondente (max)

% TaboLanceMin eh a estrutura que contem, para cada elemento de TaboLance,
% a imagem do taboleiro e o LanceMin correspondente

% MaiorTaboLanceMin eh a estrutura com o maior dos LanceMin

% POS eh a posio de MaiorTaboLanceMin em TaboLanceMin e tambem
% POS eh a posio de TaboLance em M, ou seja contem o LanceMax escolhido

minimizal([],[],Prof):-!.    % Entrada vazia retorna []
minimiza([X,J],Prof,[X,J]):-  % Quando o ganha retorna a entrada
	ganhou(X,o),!.
minimizal([C|R],[MC|MR],Prof):-
	minimiza(C,Prof,MC),   % gera um Lance Mini para a primeira estrutura
	minimizal(R,MR,Prof).  % recursao para a cauda

minimiza(X,0,X):-!.       % Quando Prof for 0 retorna a entrada
minimiza([X,J],Prof,M):-            % Gera um Lance Mini
	bagof([L,Lance],joga(o,X,Lance,L),TaboLance),
	Prof2 is Prof - 1,
	maximizal(TaboLance,TaboLanceMax,Prof2),
	menorvalor(TaboLanceMax,MenorTaboLanceMax),
	member(MenorTaboLanceMax,TaboLanceMax,POS),
	member(M,TaboLance,POS).


maximizal([],[],Prof):-!.
maximizal([C|R],[MC|MR],Prof):-
	maximiza(C,Prof,MC),    % gera um Lance Max para a primeira estrutura
	maximizal(R,MR,Prof).   % recursao para a cauda

maiorvalor([[C,J]],[C,J]):-!.
maiorvalor([[C,J]|R],[C,J]):-
	maiorvalor(R,[Taboleiro,JV]),  % sem a cabeca
	hvelha(C,HC),                  % utilidade da cabeca
	hvelha(Taboleiro,HV),          % utilidade sem a cabeca
	HC >= HV,!.                    % utilidade antiga era maior
maiorvalor([C|R],[Taboleiro,JV]):-     % escapou do corte e a cabeca perdeu
	maiorvalor(R,[Taboleiro,JV]).

menorvalor([[C,J]],[C,J]):-!.
menorvalor([[C,J]|R],[C,J]):-
	menorvalor(R,[Taboleiro,JV]), % sem a cabeca
	hvelha(C,HC),                 % utilidade da cabeca
	hvelha(Taboleiro,HV),         % utilidade sem a cabeca
	HC =< HV,!.                   % utilidade antiga era menor
menorvalor([C|R],[Taboleiro,JV]):-    % escapou do corte e a cabeca perdeu
	menorvalor(R,[V,JV]).


	
joga(X,Taboleiro,Posic,Taboleiro2):-
	possivel(Posic),              % Verifica se a jogada eh viavel
	member(b,Taboleiro,Posic),       % branco ocorre na posicao Posic de Taboleiro
	trans(Taboleiro,X,Posic,Taboleiro2).      % transforma  Taboleiro em Taboleiro2 
                                            % colocando X na posicaoo Posic
trans([C|R],X,1,[X|R]):-!.    % Se a jogada foi para a posicao 1 substitui C por X
trans([C|R],X,Posic,[C|R2]):-
	Posic2 is Posic - 1,
	trans(R,X,Posic2,R2).


% Avaliacao estatica do jogo

hv2(X,AvaliaEstat):-            % Avaliacao positiva x vence, negativa o vence
	conta(X,o,No),
	conta(X,x,Nx),
	((
	No = 3,!,                     % 3 o  No = 3        AvaliaEstat = -1000

	AvaliaEstat = -1000
	);(
	Nx = 3,!,                      % 3 x  Nx = 3        AvaliaEstat = 1000
	AvaliaEstat = 1000
	);(
	No = 0,!,                     % Nenhum o  AvaliaEstat = Nx
	AvaliaEstat = Nx
	);(
	Nx = 0,!,                     % Nenhum x  AvaliaEstat = - No
	AvaliaEstat is No * -1
	);(                            %    otherwise
	AvaliaEstat = 0
	)).
conta([],_,0):-!.                      % lista vazia tem Avaliacao = 0
conta([X|R],X,V):-                % Se X estah na cabea de uma lista a conta
	conta(R,X,V2),          % serah igual a conta da cauda mais uma unidade
	V is V2 + 1,!.
conta([_|R],X,V):-
	conta(R,X,V).

hvelha([V1,V2,V3,V4,V5,V6,V7,V8,V9],H):-
	hv2([V1,V2,V3],R1),
	hv2([V4,V5,V6],R2),
	hv2([V7,V8,V9],R3),
	hv2([V1,V4,V7],R4),
	hv2([V2,V5,V8],R5),
	hv2([V3,V6,V9],R6),
	hv2([V1,V5,V9],R7),
	hv2([V3,V5,V7],R8),
	soma([R1,R2,R3,R4,R5,R6,R7,R8],H).

soma([],0):-!.
soma([C|R],X):-
	soma(R,X2),
	X is C + X2.
